package ldaputil

import (
	"context"
	"errors"
	"net"
	"net/url"
	"time"

	"gopkg.in/ldap.v2"
)

// ConnectionPool provides a goroutine-safe pool of long-lived LDAP
// connections that will reconnect on errors.
type ConnectionPool struct {
	network string
	addr    string
	bindDN  string
	bindPw  string

	c chan *ldap.Conn
}

var defaultConnectTimeout = 5 * time.Second

func (p *ConnectionPool) connect(ctx context.Context) (*ldap.Conn, error) {
	// Dial the connection with a timeout, if the context has a
	// deadline (as it should). If the context does not have a
	// deadline, we set a default timeout.
	deadline, ok := ctx.Deadline()
	if !ok {
		deadline = time.Now().Add(defaultConnectTimeout)
	}

	c, err := net.DialTimeout(p.network, p.addr, time.Until(deadline))
	if err != nil {
		return nil, err
	}

	conn := ldap.NewConn(c, false)
	conn.Start()

	if p.bindDN != "" {
		conn.SetTimeout(time.Until(deadline))
		if _, err = conn.SimpleBind(ldap.NewSimpleBindRequest(p.bindDN, p.bindPw, nil)); err != nil {
			conn.Close()
			return nil, err
		}
	}

	return conn, err
}

// Get a fresh connection from the pool.
func (p *ConnectionPool) Get(ctx context.Context) (*ldap.Conn, error) {
	// Grab a connection from the cache, or create a new one if
	// there are no available connections.
	select {
	case conn := <-p.c:
		return conn, nil
	default:
		return p.connect(ctx)
	}
}

// Release a used connection onto the pool.
func (p *ConnectionPool) Release(conn *ldap.Conn, err error) {
	// Connections that failed should not be reused.
	if err != nil {
		conn.Close()
		return
	}

	// Return the connection to the cache, or close it if it's
	// full.
	select {
	case p.c <- conn:
	default:
		conn.Close()
	}
}

// Close all connections. Not implemented yet.
func (p *ConnectionPool) Close() {}

// Parse a LDAP URI into network and address strings suitable for
// ldap.Dial.
func parseLDAPURI(uri string) (string, string, error) {
	u, err := url.Parse(uri)
	if err != nil {
		return "", "", err
	}

	network := "tcp"
	addr := "localhost:389"
	switch u.Scheme {
	case "ldap":
		if u.Host != "" {
			addr = u.Host
		}
	case "ldapi":
		network = "unix"
		addr = u.Path
	default:
		return "", "", errors.New("unsupported scheme")
	}

	return network, addr, nil
}

// NewConnectionPool creates a new pool of LDAP connections to the
// specified server, using the provided bind credentials. The pool
// will cache at most cacheSize connections.
func NewConnectionPool(uri, bindDN, bindPw string, cacheSize int) (*ConnectionPool, error) {
	network, addr, err := parseLDAPURI(uri)
	if err != nil {
		return nil, err
	}

	return &ConnectionPool{
		c:       make(chan *ldap.Conn, cacheSize),
		network: network,
		addr:    addr,
		bindDN:  bindDN,
		bindPw:  bindPw,
	}, nil
}
